pip install cython
pip install numpy
pip install pysam
pip install https://bitbucket.org/james_taylor/bx-python/get/tip.tar.bz2
pip install biopython
pip install pybedtools
pip install matplotlib
pip install scipy
